/**
 * @file Lexer
 * @module docast-util-from-docs/lexer
 */
import type { Point } from '@flex-development/docast';
import type * as unist from 'unist';
import type { VFile } from 'vfile';
import { TokenKind } from './enums/index.mjs';
import type { Token } from './interfaces/index.mjs';
/**
 * Document tokenizer.
 *
 * @class
 */
declare class Lexer {
    /**
     * Source document.
     *
     * @public
     * @readonly
     * @instance
     * @member {string} document
     */
    readonly document: string;
    /**
     * Index of current character in {@linkcode document}.
     *
     * @protected
     * @instance
     * @member {number} offset
     */
    protected index: number;
    /**
     * List, where each index is a line number (`0`-based), and each value is the
     * number of columns in the line.
     *
     * @protected
     * @readonly
     * @instance
     * @member {number[]} indices
     */
    protected readonly indices: number[];
    /**
     * Regular expressions used to search for [comments][1] and [descendents][2].
     *
     * [1]: https://regex101.com/r/VRKolK
     * [2]: https://regex101.com/r/F1g4RP
     *
     * @protected
     * @readonly
     * @instance
     * @member {Record<'descendent' | 'parent', RegExp>} search
     */
    protected readonly search: Record<'descendent' | 'parent', RegExp>;
    /**
     * Token sequence.
     *
     * @see {@linkcode Token}
     *
     * @public
     * @readonly
     * @instance
     * @member {Token[]} tokens
     */
    readonly tokens: Token[];
    /**
     * Create a new document tokenizer.
     *
     * @param {VFile | string} source - Source document or file
     */
    constructor(source: VFile | string);
    /**
     * Check if the lexer is at the end of the document.
     *
     * @public
     *
     * @return {boolean} `true` if at end of document, `false` otherwise
     */
    get eof(): boolean;
    /**
     * Open a new token.
     *
     * @protected
     *
     * @param {TokenKind} kind - Token kind
     * @return {Token} Open token
     */
    protected enter(kind: TokenKind): Token;
    /**
     * Close an open token.
     *
     * @protected
     *
     * @param {TokenKind} kind - Token kind
     * @return {Token} Closed token
     */
    protected exit(kind: TokenKind): Token;
    /**
     * Get an offset for `point`.
     *
     * @see {@linkcode unist.Point}
     *
     * @public
     *
     * @param {unist.Point?} [point=this.point()] - Place in source document
     * @return {number} Character offset or `-1` if `point` is invalid
     */
    offset(point?: unist.Point): number;
    /**
     * Get a [*point*][1] in the document.
     *
     * **Note**: `point.column` and `point.line` will have a value of `-1` when
     * `offset` is invalid or out of bounds.
     *
     * [1]: https://github.com/syntax-tree/unist#point
     *
     * @see {@linkcode Point}
     *
     * @public
     *
     * @param {number?} [offset=this.index] - Index of character in document
     * @return {Point} Point in document
     */
    point(offset?: number): Point;
    /**
     * Change the position of the lexer.
     *
     * @protected
     *
     * @param {number} [k=1] - Difference between index of next `k`-th character
     * and current position in document
     * @return {void} Nothing
     */
    protected read(k?: number): void;
    /**
     * Tokenize the document.
     *
     * @public
     *
     * @return {void} Nothing
     */
    tokenize(): void;
}
export default Lexer;
