import {
  at,
  clamp,
  isInteger,
  isNumber,
  isString,
  reverse
} from "@flex-development/tutils";
import { ok } from "devlop";
import { TokenKind } from "./enums/index.mjs";
class Lexer {
  /**
   * Source document.
   *
   * @public
   * @readonly
   * @instance
   * @member {string} document
   */
  document;
  /**
   * Index of current character in {@linkcode document}.
   *
   * @protected
   * @instance
   * @member {number} offset
   */
  index;
  /**
   * List, where each index is a line number (`0`-based), and each value is the
   * number of columns in the line.
   *
   * @protected
   * @readonly
   * @instance
   * @member {number[]} indices
   */
  indices;
  /**
   * Regular expressions used to search for [comments][1] and [descendents][2].
   *
   * [1]: https://regex101.com/r/VRKolK
   * [2]: https://regex101.com/r/F1g4RP
   *
   * @protected
   * @readonly
   * @instance
   * @member {Record<'descendent' | 'parent', RegExp>} search
   */
  search;
  /**
   * Token sequence.
   *
   * @see {@linkcode Token}
   *
   * @public
   * @readonly
   * @instance
   * @member {Token[]} tokens
   */
  tokens;
  /**
   * Create a new document tokenizer.
   *
   * @param {VFile | string} source - Source document or file
   */
  constructor(source) {
    this.document = source = String(source);
    this.indices = [...source.matchAll(/^.*/gm)].map(([m]) => m.length + 1);
    this.index = 0;
    this.tokens = [];
    this.search = {
      descendent: /(?=(?<description>(?<=\/\*{2}(?:[\t ]+|(?:[\n\r]+[\t ]*\*[\t ]+)))(?!(?<!{)@\b\S+)[\S\s]*?(?=(?:[\t\n\r ]*\*\/)|(?:[\t\n\r *]*\*[\t ]*(?<!{)@\b\S+)))|(?<blockTag>(?<=^ +\* |(?:\/\*\* +))(?<blockTagId>@\b.+?(?=[\s}]|$))(?:\s*(?<typeExpression>{[^@].*?}?}))?(?:(?:[\n\r][\t ]*\*(?:[\n\r]|[\t ]*)(?![\t\n\r ]*\*[\t ]*(?<!{)@\b\S+|\/))|[\t ]*)(?<blockTagText>(?![\t ]*@\w+).*?)(?=(?:(?:\s*\*\n)*\s*\*\/)|(?:(?:\s*\*\n)?\s+\*[\t ]+@\w+))))|(?<inlineTag>{(?<inlineTagId>@\b.+?(?=[\s}]))[\s*]*(?<inlineTagValue>.*?)})/gms,
      parent: /(?<opener>\/\*{2}).*?(?<closer>\*\/)/gmsy
    };
  }
  /**
   * Check if the lexer is at the end of the document.
   *
   * @public
   *
   * @return {boolean} `true` if at end of document, `false` otherwise
   */
  get eof() {
    return this.index >= this.document.length;
  }
  /**
   * Open a new token.
   *
   * @protected
   *
   * @param {TokenKind} kind - Token kind
   * @return {Token} Open token
   */
  enter(kind) {
    const token = {
      end: this.point(-1),
      kind,
      start: this.point()
    };
    this.tokens.push(token);
    return token;
  }
  /**
   * Close an open token.
   *
   * @protected
   *
   * @param {TokenKind} kind - Token kind
   * @return {Token} Closed token
   */
  exit(kind) {
    const token = reverse([...this.tokens]).find((token2) => {
      return token2.kind === kind;
    });
    ok(token, "cannot exit without open token");
    ok(token.end.offset === -1, "expected open token");
    token.end = this.point();
    return token;
  }
  /**
   * Get an offset for `point`.
   *
   * @see {@linkcode unist.Point}
   *
   * @public
   *
   * @param {unist.Point?} [point=this.point()] - Place in source document
   * @return {number} Character offset or `-1` if `point` is invalid
   */
  offset(point = this.point()) {
    let found = false;
    let offset = -1;
    if (point.column && point.line && point.line <= this.indices.length && isNumber(at(this.indices, point.line - 1))) {
      for (let line = 0; line < point.line; line++) {
        for (let j = 0; j < at(this.indices, line); j++) {
          offset++;
          if (point.column === j + 1 && point.line === line + 1) {
            found = true;
            break;
          }
        }
      }
      if (!found)
        offset = -1;
    }
    return offset;
  }
  /**
   * Get a [*point*][1] in the document.
   *
   * **Note**: `point.column` and `point.line` will have a value of `-1` when
   * `offset` is invalid or out of bounds.
   *
   * [1]: https://github.com/syntax-tree/unist#point
   *
   * @see {@linkcode Point}
   *
   * @public
   *
   * @param {number?} [offset=this.index] - Index of character in document
   * @return {Point} Point in document
   */
  point(offset = this.index) {
    let index = 0;
    let line = -1;
    const point = { column: -1, line, offset };
    if (isInteger(offset) && offset > -1 && offset <= this.document.length) {
      while (++line < this.indices.length) {
        for (let j = 0; j < at(this.indices, line); j++) {
          if (index++ === offset) {
            point.column = j + 1;
            point.line = line + 1;
            break;
          }
        }
      }
    }
    return point;
  }
  /**
   * Change the position of the lexer.
   *
   * @protected
   *
   * @param {number} [k=1] - Difference between index of next `k`-th character
   * and current position in document
   * @return {void} Nothing
   */
  read(k = 1) {
    ok(isInteger(k), "expected k to be an integer");
    ok(!this.eof, "cannot read past end of document");
    this.index = clamp(this.index + k, 0, this.document.length);
    return void 0;
  }
  /**
   * Tokenize the document.
   *
   * @public
   *
   * @return {void} Nothing
   */
  tokenize() {
    while (!this.eof) {
      this.search.parent.lastIndex = this.index;
      const match = this.search.parent.exec(this.document);
      let k = 1;
      if (match) {
        const [comment] = match;
        const tokenizeDescendent = (kind, value, index) => {
          const j = comment.indexOf(value, index) - (comment.length - k);
          this.read(j);
          this.enter(kind);
          this.read(value.length);
          this.exit(kind);
          return k - (j + value.length);
        };
        k = comment.length;
        this.enter(TokenKind.COMMENT);
        for (const match2 of comment.matchAll(this.search.descendent)) {
          ok(match2.groups, "expected match.groups to be an object");
          ok(isNumber(match2.index), "expected match.index to be a number");
          const {
            description,
            blockTag,
            blockTagId,
            blockTagText,
            typeExpression,
            inlineTag,
            inlineTagId,
            inlineTagValue
          } = match2.groups;
          if (isString(blockTag)) {
            k = tokenizeDescendent(TokenKind.BLOCK_TAG, blockTag, match2.index);
            if (isString(blockTagId)) {
              k = tokenizeDescendent(
                TokenKind.BLOCK_TAG_ID,
                blockTagId,
                match2.index
              );
            }
            if (isString(typeExpression)) {
              k = tokenizeDescendent(
                TokenKind.TYPE_EXPRESSION,
                typeExpression,
                match2.index
              );
            }
            if (isString(blockTagText)) {
              k = tokenizeDescendent(
                TokenKind.BLOCK_TAG_TEXT,
                blockTagText,
                match2.index
              );
            }
          }
          if (isString(description)) {
            k = tokenizeDescendent(
              TokenKind.DESCRIPTION,
              description,
              match2.index
            );
          }
          if (isString(inlineTag)) {
            k = tokenizeDescendent(TokenKind.INLINE_TAG, inlineTag, match2.index);
            if (isString(inlineTagId)) {
              k = tokenizeDescendent(
                TokenKind.INLINE_TAG_ID,
                inlineTagId,
                match2.index
              );
            }
            if (isString(inlineTagValue)) {
              k = tokenizeDescendent(
                TokenKind.INLINE_TAG_VALUE,
                inlineTagValue,
                match2.index
              );
            }
          }
        }
        this.read(k);
        this.exit(TokenKind.COMMENT);
        continue;
      }
      this.read(k);
    }
    return void 0;
  }
}
var lexer_default = Lexer;
export {
  lexer_default as default
};
//# sourceMappingURL=lexer.mjs.map
