/**
 * @file Parser
 * @module docast-util-from-docs/parser
 */
import type { BlockTag, Comment, Content, Description, DocastNode, FlowContent, InlineTag, Position, Root } from '@flex-development/docast';
import { type Assign, type Nilable, type Optional } from '@flex-development/tutils';
import type { VFile } from 'vfile';
import { TokenKind } from './enums/index.mjs';
import type { Options, Token } from './interfaces/index.mjs';
import Lexer from './lexer.mjs';
import type { UncommentReplacer } from './types/index.mjs';
declare module 'mdast' {
    interface BreakData {
        blank?: Optional<boolean>;
        hard?: Optional<boolean>;
    }
}
declare module 'micromark-util-types' {
    interface TokenTypeMap {
        inlineTag: 'inlineTag';
    }
}
/**
 * Docblock parser.
 *
 * @class
 */
declare class Parser {
    /**
     * Document being parsed.
     *
     * @protected
     * @readonly
     * @instance
     * @member {string} document
     */
    protected readonly document: string;
    /**
     * Document tokenizer.
     *
     * @see {@linkcode Lexer}
     *
     * @protected
     * @readonly
     * @instance
     * @member {Lexer} lexer
     */
    protected readonly lexer: Lexer;
    /**
     * Parser options.
     *
     * @see {@linkcode Options}
     *
     * @protected
     * @readonly
     * @instance
     * @member {Readonly<Options>} options
     */
    protected readonly options: Readonly<Options>;
    /**
     * Syntax tree representing source file.
     *
     * @see {@linkcode Root}
     *
     * @protected
     * @readonly
     * @instance
     * @member {Root} tree
     */
    protected readonly tree: Root;
    /**
     * Create a new docblock parser.
     *
     * @see {@linkcode Options}
     * @see {@linkcode VFile}
     *
     * @param {VFile | string} source - Source document or file
     * @param {Nilable<Options>?} [options] - Parser options
     */
    constructor(source: VFile | string, options?: Nilable<Options>);
    /**
     * Get the last block tag node in the syntax tree.
     *
     * @see {@linkcode BlockTag}
     *
     * @protected
     *
     * @return {BlockTag} Last block tag node in tree
     * @throws {AssertionError} If last block tag node does not exist or is
     * invalid and `development` condition is used when importing parser module
     */
    protected get lastBlockTag(): BlockTag;
    /**
     * Get the last comment node in the syntax tree.
     *
     * @see {@linkcode Comment}
     *
     * @protected
     *
     * @return {Comment} Last comment node in tree
     * @throws {AssertionError} If last comment node does not exist or is invalid
     * and `development` condition is used when importing parser module
     */
    protected get lastComment(): Comment;
    /**
     * Get the last inline tag node in the syntax tree.
     *
     * @see {@linkcode InlineTag}
     *
     * @protected
     *
     * @return {InlineTag} Last inline tag node in tree
     * @throws {AssertionError} If last inline tag node or parent does not exist
     * and `development` condition is used when importing parser module
     */
    protected get lastInlineTag(): InlineTag;
    /**
     * Get the last inline tag parent node in the syntax tree.
     *
     * @see {@linkcode FlowContent}
     *
     * @protected
     *
     * @return {FlowContent} Last inline tag parent in tree
     * @throws {AssertionError} If last inline tag parent does not exist and
     * `development` condition is used when importing parser module
     */
    protected get lastInlineTagParent(): FlowContent;
    /**
     * Assert `token` is of `kind`.
     *
     * @protected
     *
     * @template {TokenKind} K - Token kind
     *
     * @param {K} kind - Token kind
     * @param {Token} token - Token to check
     * @return {void} Nothing; throws if `token` is invalid and `development`
     * condition is used when importing lexer module
     */
    protected assert<K extends TokenKind>(kind: K, token: Token): asserts token is Assign<Token, {
        kind: K;
    }>;
    /**
     * Handle a `blockTag` token.
     *
     * @protected
     *
     * @param {Token} token - Token to handle
     * @return {void} Nothing
     */
    protected blockTag(token: Token): void;
    /**
     * Handle a `blockTagId` token.
     *
     * @protected
     *
     * @param {Token} token - Token to handle
     * @return {void} Nothing
     * @throws {AssertionError} If block tag id cannot be extracted from source
     * file and `development` condition is used when importing parser module
     */
    protected blockTagId(token: Token): void;
    /**
     * Handle a `blockTagText` token.
     *
     * @protected
     *
     * @param {Token} token - Token to handle
     * @return {void} Nothing
     * @throws {AssertionError} If block tag text cannot be extracted from source
     * file and `development` condition is used when importing parser module
     */
    protected blockTagText(token: Token): void;
    /**
     * Handle a `comment` token.
     *
     * @protected
     *
     * @param {Token} token - Token to handle
     * @return {void} Nothing
     */
    protected comment(token: Token): void;
    /**
     * Handle a `description` token.
     *
     * @protected
     *
     * @param {Token} token - Token to handle
     * @return {void} Nothing
     */
    protected description(token: Token): void;
    /**
     * Handle an `inlineTag` token.
     *
     * @protected
     *
     * @param {Token} token - Token to handle
     * @return {void} Nothing
     */
    protected inlineTag(token: Token): void;
    /**
     * Handle an `inlineTagId` token.
     *
     * @protected
     *
     * @param {Token} token - Token to handle
     * @return {void} Nothing
     * @throws {AssertionError} If inline tag id cannot be extracted from source
     * file and `development` condition is used when importing parser module
     */
    protected inlineTagId(token: Token): void;
    /**
     * Handle an `inlineTagValue` token.
     *
     * @protected
     *
     * @param {Token} token - Token to handle
     * @return {void} Nothing
     * @throws {AssertionError} If inline tag value cannot be extracted from
     * source file and `development` condition is used when importing parser
     * module
     */
    protected inlineTagValue(token: Token): void;
    /**
     * Parse markdown.
     *
     * @see https://github.com/syntax-tree/mdast-util-from-markdown
     *
     * @protected
     *
     * @param {BlockTag | Description} node - Parent node
     * @param {string} value - Raw block tag text or raw description
     * @param {Position} position - Position of `value`
     * @return {Exclude<Content, DocastNode>[]} Markdown `node` children
     */
    protected markdown(node: BlockTag | Description, value: string, position: Position): Exclude<Content, DocastNode>[];
    /**
     * Parse the source document to an abstract syntax tree.
     *
     * @public
     *
     * @return {Root} Docblock AST
     */
    parse(): Root;
    /**
     * Handle a `typeExpression` token.
     *
     * @protected
     *
     * @param {Token} token - Token to handle
     * @return {void} Nothing
     * @throws {AssertionError} If type expression cannot be added to parent or
     * raw type expression cannot be extracted from source file and `development`
     * condition is used when importing parser module
     */
    protected typeExpression(token: Token): void;
    /**
     * Replace comment delimiter sequences in `value`.
     *
     * @see {@linkcode UncommentReplacer}
     *
     * @protected
     *
     * @param {string} value - Value to normalize
     * @param {(UncommentReplacer | string)?} [replacer=''] - A string containing
     * the text to replace for every successful match in `value`, or a function
     * that returns the replacement text
     * @return {string} `value` with comment delimiter matches replaced
     */
    protected uncomment(value: string, replacer?: UncommentReplacer | string): string;
}
export default Parser;
