import {
  at,
  constant,
  defaults,
  equal,
  fallback,
  flat,
  includes,
  isNIL,
  isNumber,
  isString,
  noop,
  sift,
  template,
  trim
} from "@flex-development/tutils";
import { ok } from "devlop";
import {
  fromMarkdown
} from "mdast-util-from-markdown";
import { codes, types } from "micromark-util-symbol";
import { u } from "unist-builder";
import { source } from "unist-util-source";
import { CONTINUE, EXIT, visit } from "unist-util-visit";
import { TokenKind } from "./enums/index.mjs";
import Lexer from "./lexer.mjs";
class Parser {
  /**
   * Document being parsed.
   *
   * @protected
   * @readonly
   * @instance
   * @member {string} document
   */
  document;
  /**
   * Document tokenizer.
   *
   * @see {@linkcode Lexer}
   *
   * @protected
   * @readonly
   * @instance
   * @member {Lexer} lexer
   */
  lexer;
  /**
   * Parser options.
   *
   * @see {@linkcode Options}
   *
   * @protected
   * @readonly
   * @instance
   * @member {Readonly<Options>} options
   */
  options;
  /**
   * Syntax tree representing source file.
   *
   * @see {@linkcode Root}
   *
   * @protected
   * @readonly
   * @instance
   * @member {Root} tree
   */
  tree;
  /**
   * Create a new docblock parser.
   *
   * @see {@linkcode Options}
   * @see {@linkcode VFile}
   *
   * @param {VFile | string} source - Source document or file
   * @param {Nilable<Options>?} [options] - Parser options
   */
  constructor(source2, options) {
    options = fallback(options, {}, isNIL);
    this.lexer = new Lexer(source2);
    this.options = Object.freeze(defaults(options, { codeblocks: "example" }));
    this.tree = u("root", { children: [] });
    this.document = this.lexer.document;
    this.lexer.tokenize();
  }
  /**
   * Get the last block tag node in the syntax tree.
   *
   * @see {@linkcode BlockTag}
   *
   * @protected
   *
   * @return {BlockTag} Last block tag node in tree
   * @throws {AssertionError} If last block tag node does not exist or is
   * invalid and `development` condition is used when importing parser module
   */
  get lastBlockTag() {
    let blockTag;
    visit(this.lastComment, "blockTag", (node) => {
      blockTag = node;
      return EXIT;
    }, true);
    ok(blockTag, "expected last block tag node");
    return blockTag;
  }
  /**
   * Get the last comment node in the syntax tree.
   *
   * @see {@linkcode Comment}
   *
   * @protected
   *
   * @return {Comment} Last comment node in tree
   * @throws {AssertionError} If last comment node does not exist or is invalid
   * and `development` condition is used when importing parser module
   */
  get lastComment() {
    let comment;
    visit(this.tree, "comment", (node) => {
      comment = node;
      return EXIT;
    }, true);
    ok(comment, "expected last comment node");
    return comment;
  }
  /**
   * Get the last inline tag node in the syntax tree.
   *
   * @see {@linkcode InlineTag}
   *
   * @protected
   *
   * @return {InlineTag} Last inline tag node in tree
   * @throws {AssertionError} If last inline tag node or parent does not exist
   * and `development` condition is used when importing parser module
   */
  get lastInlineTag() {
    let inlineTag;
    visit(this.lastInlineTagParent, "inlineTag", (node) => {
      inlineTag = node;
      return EXIT;
    }, true);
    ok(inlineTag, "expected last inline tag node");
    ok("name" in inlineTag, "expected inline tag node");
    ok("tag" in inlineTag, "expected inline tag node");
    ok("value" in inlineTag, "expected inline tag node");
    return inlineTag;
  }
  /**
   * Get the last inline tag parent node in the syntax tree.
   *
   * @see {@linkcode FlowContent}
   *
   * @protected
   *
   * @return {FlowContent} Last inline tag parent in tree
   * @throws {AssertionError} If last inline tag parent does not exist and
   * `development` condition is used when importing parser module
   */
  get lastInlineTagParent() {
    let parent;
    visit(this.lastComment, (node) => {
      switch (node.type) {
        case TokenKind.BLOCK_TAG:
        case TokenKind.DESCRIPTION:
          parent = node;
          return EXIT;
        default:
          return CONTINUE;
      }
    }, true);
    ok(parent, "expected last inline tag parent");
    return parent;
  }
  /**
   * Assert `token` is of `kind`.
   *
   * @protected
   *
   * @template {TokenKind} K - Token kind
   *
   * @param {K} kind - Token kind
   * @param {Token} token - Token to check
   * @return {void} Nothing; throws if `token` is invalid and `development`
   * condition is used when importing lexer module
   */
  assert(kind, token) {
    ok(token, "expected token");
    ok(token.kind === kind, `expected ${kind} token`);
    return void 0;
  }
  /**
   * Handle a `blockTag` token.
   *
   * @protected
   *
   * @param {Token} token - Token to handle
   * @return {void} Nothing
   */
  blockTag(token) {
    this.assert(TokenKind.BLOCK_TAG, token);
    const node = u(token.kind, {
      children: [],
      name: "",
      position: { end: token.end, start: token.start },
      tag: ""
    });
    return void this.lastComment.children.push(node);
  }
  /**
   * Handle a `blockTagId` token.
   *
   * @protected
   *
   * @param {Token} token - Token to handle
   * @return {void} Nothing
   * @throws {AssertionError} If block tag id cannot be extracted from source
   * file and `development` condition is used when importing parser module
   */
  blockTagId(token) {
    this.assert(TokenKind.BLOCK_TAG_ID, token);
    const node = this.lastBlockTag;
    node.tag = source(this.document, { end: token.end, start: token.start });
    ok(node.tag, "expected block tag id");
    node.name = node.tag.slice(1);
    return void 0;
  }
  /**
   * Handle a `blockTagText` token.
   *
   * @protected
   *
   * @param {Token} token - Token to handle
   * @return {void} Nothing
   * @throws {AssertionError} If block tag text cannot be extracted from source
   * file and `development` condition is used when importing parser module
   */
  blockTagText(token) {
    this.assert(TokenKind.BLOCK_TAG_TEXT, token);
    const node = this.lastBlockTag;
    const position = { end: token.end, start: token.start };
    const raw = source(this.document, position);
    ok(isString(raw), "expected raw block tag text");
    if (trim(raw)) {
      const children = this.markdown(
        node,
        raw,
        position
      );
      this.lastBlockTag.children.push(...children);
    }
    return void 0;
  }
  /**
   * Handle a `comment` token.
   *
   * @protected
   *
   * @param {Token} token - Token to handle
   * @return {void} Nothing
   */
  comment(token) {
    this.assert(TokenKind.COMMENT, token);
    const node = u(token.kind, {
      children: [],
      code: null,
      position: { end: token.end, start: token.start }
    });
    return void this.tree.children.push(node);
  }
  /**
   * Handle a `description` token.
   *
   * @protected
   *
   * @param {Token} token - Token to handle
   * @return {void} Nothing
   */
  description(token) {
    this.assert(TokenKind.DESCRIPTION, token);
    const position = { end: token.end, start: token.start };
    const raw = source(this.document, position);
    ok(isString(raw), "expected raw description");
    const node = u(token.kind, {
      children: [],
      position
    });
    node.children = this.markdown(node, raw, position);
    return void this.lastComment.children.push(node);
  }
  /**
   * Handle an `inlineTag` token.
   *
   * @protected
   *
   * @param {Token} token - Token to handle
   * @return {void} Nothing
   */
  inlineTag(token) {
    this.assert(TokenKind.INLINE_TAG, token);
    const position = { end: token.end, start: token.start };
    return void visit(this.lastInlineTagParent, token.kind, (n, index, parent) => {
      if (!equal(n.position, position))
        return CONTINUE;
      ok(isNumber(index), "expected indexed node");
      ok(parent, "expected parent");
      parent.children.splice(index, 1, u(token.kind, {
        name: "",
        position,
        tag: "",
        value: ""
      }));
      return EXIT;
    });
  }
  /**
   * Handle an `inlineTagId` token.
   *
   * @protected
   *
   * @param {Token} token - Token to handle
   * @return {void} Nothing
   * @throws {AssertionError} If inline tag id cannot be extracted from source
   * file and `development` condition is used when importing parser module
   */
  inlineTagId(token) {
    this.assert(TokenKind.INLINE_TAG_ID, token);
    const node = this.lastInlineTag;
    node.tag = source(this.document, { end: token.end, start: token.start });
    ok(node.tag, "expected inline tag id");
    node.name = node.tag.slice(1);
    return void 0;
  }
  /**
   * Handle an `inlineTagValue` token.
   *
   * @protected
   *
   * @param {Token} token - Token to handle
   * @return {void} Nothing
   * @throws {AssertionError} If inline tag value cannot be extracted from
   * source file and `development` condition is used when importing parser
   * module
   */
  inlineTagValue(token) {
    this.assert(TokenKind.INLINE_TAG_VALUE, token);
    const node = this.lastInlineTag;
    node.value = source(this.document, { end: token.end, start: token.start });
    ok(node.value, "expected inline tag value");
    return void 0;
  }
  /**
   * Parse markdown.
   *
   * @see https://github.com/syntax-tree/mdast-util-from-markdown
   *
   * @protected
   *
   * @param {BlockTag | Description} node - Parent node
   * @param {string} value - Raw block tag text or raw description
   * @param {Position} position - Position of `value`
   * @return {Exclude<Content, DocastNode>[]} Markdown `node` children
   */
  markdown(node, value, position) {
    const codeblocks = flat(sift([
      this.options.codeblocks
    ]));
    const codeblock = "tag" in node && codeblocks.some((check) => {
      return isString(check) ? includes([node.name, node.tag], check) : [node.name, node.tag].some((str) => check.test(str));
    }) && !value.startsWith("```");
    const columns = [];
    value = this.uncomment(value, (match) => {
      columns.push(columns.length ? match.length : position.start.column - 1);
      return "";
    });
    if (codeblock)
      value = template("```\n{value}\n```", { value });
    const tree = fromMarkdown(value, {
      extensions: [
        {
          text: {
            [codes.leftCurlyBrace]: {
              /**
               * Construct name.
               */
              name: TokenKind.INLINE_TAG,
              /**
               * Guard whether `code` can come before this construct.
               *
               * @see {@linkcode CharacterCode}
               *
               * @this {TokenizeContext}
               *
               * @param {CharacterCode} code - Previous character code
               * @return {boolean} `true` if `code` allowed before construct
               */
              previous(code) {
                return code !== codes.backslash;
              },
              /**
               * Set up a state machine to process character codes.
               *
               * @see {@linkcode Effects}
               * @see {@linkcode State}
               *
               * @this {TokenizeContext}
               *
               * @param {Effects} effects - Context object to transition state
               * @param {State} ok - Success state function
               * @param {State} nok - Error state function
               * @return {State} Initial state
               */
              tokenize(effects, ok2, nok) {
                const inside = (code) => {
                  effects.consume(code);
                  if (code === codes.rightCurlyBrace && this.previous !== codes.backslash) {
                    effects.exit(TokenKind.INLINE_TAG);
                    return ok2;
                  }
                  return inside;
                };
                function begin(code) {
                  if (code !== codes.atSign) {
                    return nok(code);
                  }
                  effects.consume(code);
                  return inside;
                }
                function start(code) {
                  effects.enter(TokenKind.INLINE_TAG);
                  effects.consume(code);
                  return begin;
                }
                return start;
              }
            }
          }
        },
        ...fallback(this.options.micromarkExtensions, [], isNIL)
      ],
      mdastExtensions: [
        {
          enter: {
            [types.atxHeading]: noop,
            /**
             * Enter a hard break.
             *
             * @this {CompileContext}
             *
             * @param {MToken} token - Micromark token
             * @return {void} Nothing
             */
            [types.hardBreakEscape](token) {
              return void this.enter({
                data: { hard: true },
                type: "break"
              }, token);
            },
            /**
             * Enter an inline tag.
             *
             * @this {CompileContext}
             *
             * @param {MToken} token - Micromark token
             * @return {void} Nothing
             */
            [TokenKind.INLINE_TAG](token) {
              return void this.enter({
                type: TokenKind.INLINE_TAG
              }, token);
            },
            /**
             * Enter a blank line.
             *
             * @this {CompileContext}
             *
             * @param {MToken} token - Micromark token
             * @return {void} Nothing
             */
            [types.lineEndingBlank](token) {
              return void this.enter({
                data: { blank: true },
                type: "break"
              }, token);
            },
            [types.setextHeading]: noop,
            [types.setextHeadingLineSequence]: noop,
            [types.setextHeadingText]: noop
          },
          exit: {
            [types.atxHeading]: noop,
            /**
             * Exit a hard break.
             *
             * @this {CompileContext}
             *
             * @param {MToken} token - Micromark token
             * @return {void} Nothing
             */
            [types.hardBreakEscape](token) {
              return void this.exit(token);
            },
            /**
             * Exit an inline tag.
             *
             * @this {CompileContext}
             *
             * @param {MToken} token - Micromark token
             * @return {void} Nothing
             */
            [TokenKind.INLINE_TAG](token) {
              return void this.exit(token);
            },
            /**
             * Exit a blank line.
             *
             * @this {CompileContext}
             *
             * @param {MToken} token - Micromark token
             * @return {void} Nothing
             */
            [types.lineEndingBlank](token) {
              return void this.exit(token);
            },
            [types.setextHeading]: noop,
            [types.setextHeadingLineSequence]: noop,
            [types.setextHeadingText]: noop
          },
          transforms: [
            /**
             * Node position transformer.
             *
             * @see {@linkcode MdastRoot}
             *
             * @param {MdastRoot} tree - Markdown syntax tree
             * @return {void} Nothing
             */
            (tree2) => {
              return void visit(tree2, (m) => {
                ok(m.position, "expected non-generated node");
                const span = m.position.end.line - m.position.start.line;
                for (const [i, removed] of columns.entries()) {
                  const line = i + 1;
                  if (m.position.start.line === line) {
                    m.position.start.column += removed;
                  }
                  if (m.position.end.line === line && m.type !== "break") {
                    m.position.end.column += removed;
                  }
                }
                m.position.start.line += position.start.line - 1;
                m.position.start.offset = this.lexer.offset(m.position.start);
                if (codeblock && m.type === "code") {
                  m.position.end.column = position.end.column;
                  m.position.end.line = position.end.line;
                  m.position.end.offset = position.end.offset;
                } else {
                  if (m.type === "break") {
                    m.position.end.column = 1;
                    if (m.data?.hard) {
                      m.position.end.column += m.position.start.column;
                    }
                  }
                  m.position.end.line = m.position.start.line + span;
                  m.position.end.offset = this.lexer.offset(m.position.end);
                }
                return CONTINUE;
              });
            },
            /**
             * Insert new `break` nodes before `break` nodes that mark the end
             * of an empty line.
             *
             * @see {@linkcode MdastRoot}
             *
             * @param {MdastRoot} tree - Markdown syntax tree
             * @return {void} Nothing
             */
            (tree2) => {
              return void visit(tree2, "break", (m, index, parent) => {
                if (m.data?.blank) {
                  ok(m.position, "expected non-generated node");
                  ok(isNumber(m.position.start.offset), "expected start offset");
                  ok(isNumber(index), "expected indexed node");
                  ok(parent, "expected parent");
                  const { start } = m.position;
                  const offset = start.offset - start.column;
                  parent.children.splice(index, 1, {
                    position: {
                      end: this.lexer.point(offset + 1),
                      start: this.lexer.point(offset)
                    },
                    type: m.type
                  }, m);
                  return index + 2;
                }
                return CONTINUE;
              });
            },
            /**
             * Replace or remove newline characters in `text` nodes.
             *
             * Newline characters will be replaced if with spaces if they denote
             * the beginning a comment delimiter sequence sandwiched between
             * description or block tag content. They will be removed if they
             * are preceded by a hard break.
             *
             * @see {@linkcode MdastRoot}
             *
             * @param {MdastRoot} tree - Markdown syntax tree
             * @return {void} Nothing
             */
            (tree2) => {
              return void visit(tree2, "text", (m, index, parent) => {
                ok(isString(m.value), "expected string value");
                const search = /[\n\r]/g;
                if (search.test(m.value)) {
                  ok(m.position, "expected non-generated node");
                  ok(isNumber(m.position.start.offset), "expected start offset");
                  ok(isNumber(index), "expected indexed node");
                  ok(parent, "expected parent");
                  const prev = at(
                    parent.children,
                    index - 1
                  );
                  if (prev?.type === "break" && prev.data?.hard) {
                    ok(/^[\n\r]/.test(m.value), "expected newline start");
                    let { offset } = m.position.start;
                    m.value = m.value.slice(1);
                    offset = this.document.indexOf(m.value, offset);
                    m.position.start = this.lexer.point(offset);
                  } else {
                    if (/[\n\r]$/.test(m.value)) {
                      m.position.end.column = 1;
                      m.position.end.offset = this.lexer.offset(m.position.end);
                    }
                    m.value = m.value.replaceAll(search, " ");
                  }
                }
                return CONTINUE;
              });
            }
          ]
        },
        ...fallback(this.options.mdastExtensions, [], isNIL)
      ]
    });
    return tree.children;
  }
  /**
   * Parse the source document to an abstract syntax tree.
   *
   * @public
   *
   * @return {Root} Docblock AST
   */
  parse() {
    const transforms = fallback(this.options.transforms, [], isNIL);
    const paragraphs = (tree) => {
      return void visit(tree, "paragraph", (m, index, parent) => {
        ok(isNumber(index), "expected indexed node");
        ok(parent, "expected parent");
        switch (parent.type) {
          case "blockTag":
          case "listItem":
            parent.children.splice(index, 1, ...m.children);
            return index;
          default:
            return CONTINUE;
        }
      });
    };
    for (const token of this.lexer.tokens)
      void this[token.kind](token);
    transforms.unshift(paragraphs);
    for (const transform of transforms)
      void transform(this.tree);
    return this.tree;
  }
  /**
   * Handle a `typeExpression` token.
   *
   * @protected
   *
   * @param {Token} token - Token to handle
   * @return {void} Nothing
   * @throws {AssertionError} If type expression cannot be added to parent or
   * raw type expression cannot be extracted from source file and `development`
   * condition is used when importing parser module
   */
  typeExpression(token) {
    this.assert(TokenKind.TYPE_EXPRESSION, token);
    ok(
      this.lastBlockTag.children.every((child) => child.type !== token.kind),
      "blockTag may only contain one typeExpression node"
    );
    const position = { end: token.end, start: token.start };
    const raw = source(this.document, position);
    ok(raw, "expected raw type expression");
    const node = u(token.kind, {
      position,
      value: this.uncomment(raw.slice(1, -1))
    });
    return void this.lastBlockTag.children.unshift(node);
  }
  /**
   * Replace comment delimiter sequences in `value`.
   *
   * @see {@linkcode UncommentReplacer}
   *
   * @protected
   *
   * @param {string} value - Value to normalize
   * @param {(UncommentReplacer | string)?} [replacer=''] - A string containing
   * the text to replace for every successful match in `value`, or a function
   * that returns the replacement text
   * @return {string} `value` with comment delimiter matches replaced
   */
  uncomment(value, replacer = "") {
    return value.replaceAll(
      /^(?:[\t ]*\*[\t ]{0,2})?/gm,
      isString(replacer) ? constant(replacer) : replacer
    );
  }
}
var parser_default = Parser;
export {
  parser_default as default
};
//# sourceMappingURL=parser.mjs.map
